//---------------------------------------------------------------------------------------------------------------------------------------------------
// Last Change: 2023-01-25      Name: export_.cpp
//---------------------------------------------------------------------------------------------------------------------------------------------------
// This file contains the language-independent implementation of the plug-in EXPORT_.DLL. All texts and resources that are language-dependent are
// located in an additional EXPORT.DLL, of which sources can be found in the subdirectories \E (for English) and \D (for German).
//---------------------------------------------------------------------------------------------------------------------------------------------------

#define MK_PRODUCT_REGKEY L"EXPORT"

//---------------------------------------------------------------------------------------------------------------------------------------------------

#include  "windows.h"
#include  "windowsx.h"
#include  "stdlib.h"
#include  "stdio.h"
#include  "math.h"

#ifdef _MSC_VER
#pragma warning( disable:4458 ) // declaration hides class member
#endif

#include  "gdiplus.h"

#ifdef _MSC_VER
#pragma warning( default:4458 ) // declaration hides class member
#endif

#include  "d2d1.h"
#include  "d2d1_1.h"
#include  "dwrite.h"
#include  "wincodec.h"

#include  "x:\mkcad6\mki_lib6\mki_lib6.h"
#include  "x:\mkcad6\mki_lib6\CompileDate.h"
#include  "x:\mkcad6\mki_lib6\Dialog.h"

//--------------------------------------------------------------------------------------------------------------------------------------------------
// Language-dependent texts in EXPORT.DLL

constexpr LPSTR TXT_ID_HEADER = "Malz++Kassner TXT 6.00";

MKI_INTERFACE_IMPORT LPWSTR
  eStartUpText[],
  eDialogText[];

//---------------------------------------------------------------------------------------------------------------------------------------------------

class CGlobalData
{

  public:

  CGlobalData( void )
  {
    ZeroMemory( this, sizeof( *this ) );
  }

  HINSTANCE m_hInstDLL;     // Instance handle of the main DLL
  HINSTANCE m_hLanguage;    // Instance handle of the language DLL
  HINSTANCE m_hGlobalInst;  // Instance handle of the serving application
  HWND      m_hGlobalWnd;   // Main window handle of the serving application

  MKI_BITMAP_PTR m_pBitmap; // Handle of command icon bitmap

  MKI_POINT m_cOrigin;  // Current origin
  __int32   m_nError;   // Current error status ( 0 = OK )
};

CGlobalData gGlobalData;


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is used as callback procedure for an object enumeration based on the interface procedure MKI_EnumerateAll(). It simply runs through
// of the object described in f_pEnumData and writes the coordinates of all its definition points to the currently open file. After each coordinate
// pair written, it also updates the current progress indicator and checks whether the user has pressed the Cancel button to stop the export.
bool
PlugInEnumCallback( MKI_CONST_ENUMDEF_DATA_PTR f_pEnumData, void* f_pUserData )
{
  f_pUserData;  // Touch formal parameter to avoid C4100

  MKI_DUMMYSTRW szText1, 
                szText2;

  switch( f_pEnumData->m_nEnumData )
  {
    case MKI_ENUMDATA_CURVE:
    case MKI_ENUMDATA_SURFACE:
    case MKI_ENUMDATA_MARK:
      for( __int32 nCount = 0; nCount < f_pEnumData->m_nEnumCount; nCount++ )
      {
        MKI_FileWriteInt16( MKI_DB_POINT_ANY );
        MKI_FileWriteComma(); 
        MKI_FileWriteDouble( f_pEnumData->m_pPointData[nCount].m_cPoint.x - gGlobalData.m_cOrigin.x );
        MKI_FileWriteComma(); 
        MKI_FileWriteDouble( f_pEnumData->m_pPointData[nCount].m_cPoint.y - gGlobalData.m_cOrigin.y );
        MKI_FileWriteSemi();

        if( MKI_FileWriteError() )
        {
          gGlobalData.m_nError = 1;
          break;
        }

        if( MKI_DialogProgressIsPending() )
        {
          MKI_ConvertNumberString( szText2, MKI_FileWriteCurrentLine() );
          MKI_PrintW( szText1, eDialogText[3], szText2 );
          MKI_ConvertBytesString( szText2, MKI_FileWriteCurrentSize() );
          MKI_DialogUpdateProgress( szText1, szText2, 0, 0 );

          if( MKI_DialogIsCancelled() )
          {
            gGlobalData.m_nError = 999;
            break;
          }
        }
      }
      break;
  }
  return( !gGlobalData.m_nError );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure performs the export. It starts the object enumeration (which writes coordinate pairs to the file).
// In addition, it initializes and display a progress indicator window to inform the user about the current progress and to allow him to cancel the
// export. Since the final file size is not known in advance, the progress indicator does not include a percent bar.
void
PlugInExport( HANDLE f_hFileHandle, const LPCWSTR f_pszFileName, __int32 f_nFlag )
{
  MKI_FILENAMEW szFileName;
  MKI_DUMMYSTRW szDummyStr;
  MKI_SYSTEMDEF cSystemDef;

  gGlobalData.m_nError = 0;

  if( MKI_SystemGetCurrent( cSystemDef ) >= 0 )
  {
    gGlobalData.m_cOrigin = cSystemDef.m_cSystemOrigin;
  }
  else
  {
    gGlobalData.m_cOrigin.SetZero();
  }

  if( !MKI_FileWriteInitDisk( f_hFileHandle, false, false ) )
  {
    gGlobalData.m_nError = 1;
    return;
  }

  MKI_FileWriteZeros( true );

  MKI_FileBaseName( f_pszFileName, szFileName );
  MKI_PrintW( szDummyStr, eDialogText[2], szFileName );
  MKI_DialogShowProgress( eDialogText[0], szDummyStr, false );

  MKI_FileWriteTextData( TXT_ID_HEADER );
  MKI_FileWriteSemi();

  MKI_EnumerateAll( f_nFlag, f_nFlag,
                    MKI_ENUM_PLAIN | MKI_ENUM_LINES | MKI_ENUM_LINETYPES,
                    (MKI_ENUMOBJECT_PROC) PlugInEnumCallback, nullptr );

  MKI_FileWriteInt16( MKI_DB_END );
  MKI_FileWriteSemi();

  if( !MKI_FileWriteFlush() )
    gGlobalData.m_nError = 1;

  MKI_FileWriteExit();
  MKI_DialogHideProgress();
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This DLL entry procedure must exist in any DLL to be used in Win32. Since our DLL does all necessary initialization in its MKI_PlugInInit()
// procedure, this procedure is quite empty.
BOOL WINAPI
DllMain( HINSTANCE f_hInstance, DWORD f_dwReason, LPVOID f_pDummy )
{
  f_pDummy; // Touch formal parameter to avoid C4100

  switch( f_dwReason )
  {
    case DLL_PROCESS_ATTACH:
      gGlobalData.m_hInstDLL = f_hInstance;
      break;

    case DLL_PROCESS_DETACH:
      gGlobalData.m_hInstDLL = nullptr;
      break;
  }
  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the plug-in is loaded by the serving application. Its main tasks are:
// - Checking whether it is compatible with the given f_nInterfaceVersion
// - Checking whether it is licensed to the given serial number (optional)
// - Storing of the serving application's instance and main windows handle for further use
// - Loading of the language-dependent library
// - Filling in the plug-in ID structure of which address is passed in f_pPlugInID
// - Loading of options from the registry database
// - Loading profiles
// - Allocating any static memory required
MKI_INTERFACE_EXPORT __int32
MKI_PlugInInit( const LPCSTR f_pszSerialNumber, HINSTANCE f_hMainInst, HWND f_hMainWnd,
                __int32 f_nInterfaceVersion, MKI_PLUGIN_ID_REF f_rPlugInID )
{
  MKI_FILENAMEW szFileName;

  if( f_nInterfaceVersion > MKI_INTERFACE_VERSION )
    return( MKI_INIT_OLD_PLUGIN );
  else if( f_nInterfaceVersion < MKI_INTERFACE_VERSION )
    return( MKI_INIT_OLD_APPLICATION );

  if( MKI_EqualA( f_pszSerialNumber, "DEMO" ) )
    return( MKI_INIT_ERROR );

  gGlobalData.m_hGlobalInst = f_hMainInst;
  gGlobalData.m_hGlobalWnd  = f_hMainWnd;

  gGlobalData.m_hLanguage = LoadLibraryW( MK_PRODUCT_REGKEY L".DLL" );
  MKI_FileApplicationPath( MK_PRODUCT_REGKEY L".PNG", szFileName );
  gGlobalData.m_pBitmap = MKI_BitmapLoadImage( szFileName );

  // Fill the plug-in information data block
  f_rPlugInID.m_nOwnerID    = MKI_DB_OWNER_MK;
  f_rPlugInID.m_nPlugInID   = 10000;
  f_rPlugInID.m_nPlugInCTRL = MKI_PLUGINCTRL_STARTER | MKI_PLUGINCTRL_PRO | MKI_PLUGINCTRL_TEST | MKI_PLUGINCTRL_DEMO;
  f_rPlugInID.m_pContainer  = nullptr;

  f_rPlugInID.m_cPlugInProc.m_pfnInputPointInitProc = (MKI_INPUTPOINTINIT_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputPointMoveProc = (MKI_INPUTPOINTMOVE_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputPointExitProc = (MKI_INPUTPOINTEXIT_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputDisplayProc   = (MKI_INPUTDISPLAY_PROC)   nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputParameterProc = (MKI_INPUTPARAMETER_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputCancelProc    = (MKI_INPUTCANCEL_PROC)    nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputFinishProc    = (MKI_INPUTFINISH_PROC)    nullptr;

  f_rPlugInID.m_cPlugInData.m_nType = MKI_PLUGINMENU_EXPORT;

  f_rPlugInID.m_cPlugInData.m_cInputData.m_nCommandMode = MKI_COMMAND_DIRECT;
  f_rPlugInID.m_cPlugInData.m_cMenuData.m_pszMenuEntry   = eStartUpText[1];
  f_rPlugInID.m_cPlugInData.m_cMenuData.m_pszDescription = eStartUpText[2];

  f_rPlugInID.m_cPlugInData.m_cIconData.m_pIconBitmap  = gGlobalData.m_pBitmap;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconXOffset = MKI_ICON_COMMAND_FRAME;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconYOffset = MKI_ICON_COMMAND_FRAME;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconXSize   = MKI_ICON_COMMAND_SIZE;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconYSize   = MKI_ICON_COMMAND_SIZE;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconMode    = 0;

  f_rPlugInID.m_pCommandData  = nullptr;
  f_rPlugInID.m_pButtonData   = nullptr;
  f_rPlugInID.m_pExtendedData = nullptr;

  f_rPlugInID.m_pszFileType      = eStartUpText[3];
  f_rPlugInID.m_pszFileExtension = eStartUpText[4];
  f_rPlugInID.m_pszFileDefault   = eStartUpText[5];

  return( MKI_INIT_OK );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the plug-in is removed by the serving application. Its main tasks are:
// - Checking whether anything is to be saved. If so, it should display a message information the user
//   about it and allowing him to save those changes.
// - Freeing of all statically allocated memory.
// If this procedure returns false, the serving application will not be able to terminate. So please, do only return false if shutting down the
// plug-in now would severely damage or destroy user data.
MKI_INTERFACE_EXPORT bool
MKI_PlugInExit( bool f_fSaveSettings )
{
  f_fSaveSettings;  // Touch formal parameter to avoid C4100

  MKI_BitmapFreeImage( gGlobalData.m_pBitmap );

  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the user wants to save a non-MKD drawing using the File>Save as command. For an export filter, its task is simply
// to save the file.
MKI_INTERFACE_EXPORT bool
MKI_PlugInExport( const LPCWSTR f_pszFileName, __int32 f_nFilterMode, __int32 f_nFlag, 
                  const LPCWSTR f_pszCollectionName, const LPCWSTR f_pszItemName, void* f_pData )
{
  f_pData;              // Touch formal parameter to avoid C4100
  f_nFilterMode;        // Touch formal parameter to avoid C4100
  f_pszCollectionName;  // Touch formal parameter to avoid C4100
  f_pszItemName;        // Touch formal parameter to avoid C4100

  HANDLE  hFileHandle;
  bool    fResult = false;

  if( MKI_FileCreate( &hFileHandle, f_pszFileName ) )
  {
    PlugInExport( hFileHandle, f_pszFileName, f_nFlag );

    MKI_FileClose( hFileHandle );
    switch( gGlobalData.m_nError )
    {
      case 999:
        MKI_FileDelete( f_pszFileName );
        MKI_DialogMessage( eDialogText[4], eDialogText[0], MKI_MESSAGE_OK );
        fResult = true;
        break;

      case 0:
        fResult = true;
        break;

      default:
        MKI_FileDelete( f_pszFileName );
        MKI_DialogMessage( eDialogText[5], eDialogText[0], MKI_MESSAGE_OK );
        fResult = false;
        break;
    }
  }
  return( fResult );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when a module's command is chosen by the user. For an export filter, its main tasks are:
// - Parameter editing
// - Help file display
MKI_INTERFACE_EXPORT __int32
MKI_PlugInCommand( __int32 f_nCommandID, __int32 f_nExecMode, void* f_pData )
{
  f_nCommandID; // Touch formal parameter to avoid C4100
  f_pData;      // Touch formal parameter to avoid C4100

  // Check which type of action is requested
  switch( f_nExecMode )
  {
    // Display the corresponding help topic
    case MKI_EXECMODE_HELP:
      MKI_DialogHelpTopic( gGlobalData.m_hGlobalWnd, MK_PRODUCT_REGKEY, 0 );
      return( true );

    // Edit the options
    case MKI_EXECMODE_USER:
      {
        MKI_PLUGIN_CONTAINER  cData;

        MKI_CopyW( cData.m_szRegistry, MK_PRODUCT_REGKEY );
        MKI_GetBuildDate( __DATE__, cData.m_szBuildDate );
        cData.m_pszCaption   = eDialogText[0];
        cData.m_pszVendor    = nullptr;
        cData.m_pszCopyright = eStartUpText[0];
        cData.m_nUserMode    = MKI_PLUGINMODE_FREE;

        MKI_DialogPlugInAbout( GetActiveWindow(), cData, nullptr );
      }
      return( true );

    // Execute the command
    case MKI_EXECMODE_SYSTEM:
      // Nothing to do here...
      break;

    // Read drawing-dependent data
    case MKI_EXECMODE_GET_PROFILE:
      // Nothing to do here...
      break;

    // Write drawing-dependent data
    case MKI_EXECMODE_SET_PROFILE:
      // Nothing to do here...
      break;

    // Undo selected by user
    case MKI_EXECMODE_UNDO:
      // Nothing to do here...
      break;

    // Redo selected by user
    case MKI_EXECMODE_REDO:
      // Nothing to do here...
      break;

    // Clear undo data
    case MKI_EXECMODE_CLEAR_UNDO:
      // Nothing to do here...
      break;
  }
  return( false );
}
