//---------------------------------------------------------------------------------------------------------------------------------------------------
// Last Change: 2024-03-05      Name: sample_.cpp
//---------------------------------------------------------------------------------------------------------------------------------------------------
// This file contains the language-independent implementation of the plug-in SAMPLE_.DLL. All texts and resources that are language-dependent are
// located in an additional SAMPLE.DLL, of which sources can be found in the subdirectories \E (for English) and \D (for German).
//---------------------------------------------------------------------------------------------------------------------------------------------------

#define MK_PRODUCT_REGKEY L"SAMPLE"

//---------------------------------------------------------------------------------------------------------------------------------------------------

#include  "windows.h"
#include  "windowsx.h"
#include  "stdlib.h"
#include  "stdio.h"
#include  "math.h"

#ifdef _MSC_VER
#pragma warning( disable:4458 ) // declaration hides class member
#endif

#include  "gdiplus.h"

#ifdef _MSC_VER
#pragma warning( default:4458 ) // declaration hides class member
#endif

#include  "d2d1.h"
#include  "d2d1_1.h"
#include  "dwrite.h"
#include  "wincodec.h"

#include  "x:\mkcad6\mki_lib6\mki_lib6.h"
#include  "x:\mkcad6\mki_lib6\CompileDate.h"
#include  "x:\mkcad6\mki_lib6\Dialog.h"

//--------------------------------------------------------------------------------------------------------------------------------------------------

constexpr __int32 SAMPLE_PLUGIN_ID      = 12000;
constexpr __int32 SAMPLE_PLUGIN_VERSION = 650;

constexpr LPSTR SAMPLE_PROFILE_VERSION = "2016.0";

//--------------------------------------------------------------------------------------------------------------------------------------------------

constexpr __int32 STAR_INDENT_MIN = 3;    // Minimum number of indents
constexpr __int32 STAR_INDENT_MAX = 100;  // Maximum number of indents

//--------------------------------------------------------------------------------------------------------------------------------------------------

enum
{
  SAMPLE_MENU_POLYGON  = 1,
  SAMPLE_MENU_EXTENDED = 2,
  SAMPLE_MENU_ABOUT    = 4,
};

//--------------------------------------------------------------------------------------------------------------------------------------------------
// Language-dependent texts in SAMPLE.DLL

MKI_INTERFACE_IMPORT LPWSTR
  eStartUpText[],
  eDialogText[],
  eCommandName[],
  eCommandEntry[],
  eNewPoint[];

MKI_INTERFACE_IMPORT MKI_TOOLTIP_ENTRY
  eTooltipDialogProperty[];


//--------------------------------------------------------------------------------------------------------------------------------------------------
class CPropertyData : public MKI_MULTI_BASE
{

public:

  enum
  {
    STAR_USE_INDENTNUM = 0x00000001,  // Property usage flags
  };

  void
  Fill( __int32 f_nIndentNum )
  {
    MKI_MULTI_BASE::Handle( m_nIndentNum, f_nIndentNum, STAR_USE_INDENTNUM );
  }

  void
  Apply( __int32& f_rnIndentNum )
  {
    if( m_nUse & STAR_USE_INDENTNUM )
      f_rnIndentNum = m_nIndentNum;
  }

  __int32 m_nIndentNum;   // STAR_USE_INDENTNUM
};


//--------------------------------------------------------------------------------------------------------------------------------------------------
class CGlobalData
{

public:

  CGlobalData( void )
  {
    ZeroMemory( this, sizeof( *this ) );
  }
  
  HINSTANCE m_hInstDLL;     // Instance handle of the main DLL
  HINSTANCE m_hLanguage;    // Instance handle of the language DLL
  HINSTANCE m_hGlobalInst;  // Instance handle of the serving application
  HWND      m_hGlobalWnd;   // Main window handle of the serving application
  
  MKI_BITMAP_PTR  m_pBitmap;    // Handle of command icon bitmap

  MKI_POINT       m_acPoint[4]; // Points entered by the user
  __int32         m_nCommandID; // Current Command's ID
  CPropertyData   m_cProperty;  // Property buffer for property editing
  
  MKI_COMMAND_DATA  m_acCommandData[MKI_PLUGIN_SUBMENU_MAX];  // Command data
  MKI_EXTENDED_DATA m_acExtendedData[2];                      // Extended unit data
};


//--------------------------------------------------------------------------------------------------------------------------------------------------
class CDialogData
{

public:

  CDialogData( void )
  {
    ZeroMemory( this, sizeof( *this ) );
  }
  
  MKI_DUMMYSTRW     m_szCaption;     // Current dialog window caption
  HWND              m_hOldFocus;     // Previous focus to be restored
  bool              m_fUserUpdate;   // Indicates if change shall be processed
  __int32           m_nExitDialog;   // Button with which the dialog was cancelled
  CPropertyData     m_cCopyProperty; // Temporary copy of Property
  MKI_TOOLTIP_LIST  m_cTooltip;      // Tooltip data
  MKI_SUBCLASS_LIST m_cSubclassList; // Control subclassing
};

CGlobalData   gGlobalData;
CDialogData   gDialogData;
CPropertyData gUserData;


//--------------------------------------------------------------------------------------------------------------------------------------------------
// Command Description
//
// The structure MKI_COMMAND_ENTRY contains a command description. It lists the points to be entered by the user and their types. Each point
// description consists of three values: The point type (MKI_POINT_???), the index of "related point", i.e. the point to which this point is
// relative, and a textual description of that point. The following command required exactly 4 points to be entered (MKI_COMMAND_FIXED = number
// of points to be entered is known in advance, "4" says that this known number is three).
// Following a description of those three points to be entered:
// 1) MKI_POINT_CENTER = center of a circle, ellipse, etc.
//    "-1"         = no "related point"
//    nullptr      = default description
// 2) MKI_POINT_RADIUS = radius of a circle
//    "0"          = relative to the point with the index zero (the center point)
//    nullptr      = default description
// 3) MKI_POINT_RADIUS = radius of a circle
//    "0"          = relative to the point with the index zero (the center point)
//    nullptr      = default description
// 4) MKI_POINT_ANGLE = rotation angle
//    "0"          = relative to the point with the index zero (the center point)
//    nullptr      = default description
// Please note that the description strings of the points 2) and 3) will be overwritten before passed to the serving application in order to
// change the description to "Enter Radius 1" and "Enter Radius 2" (see end of procedure MKI_PlugInInit()).

// Command description structure
MKI_COMMAND_ENTRY gPointInfo[] =
{
  {
    MKI_COMMAND_FIXED,
    4,
    { MKI_POINT_CIRCLE_CENTER, MKI_POINT_RADIUS, MKI_POINT_RADIUS, MKI_POINT_ANGLE },
    { -1, 0, 0, 0 },
    { nullptr, nullptr, nullptr, nullptr }
  }
};


//---------------------------------------------------------------------------------------------------------------------------------------------------
// Initializes the plug-in's settings from the registry database.
void
PlugInInitSettings( void )
{
  gUserData.m_nIndentNum = 6;
  gUserData.SetNone();
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// Loads the plug-in's settings from the registry database.
void
PlugInLoadSettings( void )
{
  PlugInInitSettings();

  if( MKI_ProfileReadKeyOpen( MK_PRODUCT_REGKEY ) )
  {
    MKI_ProfileReadInt32( L"Indent", gUserData.m_nIndentNum );
    MKI_ProfileReadKeyClose();
  }
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// Saves the plug-in's settings to the registry database.
void
PlugInSaveSettings( void )
{
  if( MKI_ProfileWriteKeyOpen( MK_PRODUCT_REGKEY ) )
  {
    MKI_ProfileWriteClear();
    MKI_ProfileWriteString( L"Version", SAMPLE_PROFILE_VERSION );
    MKI_ProfileWriteInt32 ( L"Indent", gUserData.m_nIndentNum );
    MKI_ProfileWriteKeyClose();
  }
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// Callback procedure for dialog box handling of DialogEditParameters(). This procedure handles all messages sent to the dialog window.
INT_PTR CALLBACK
DialogEditParametersManage( HWND f_hDlg, UINT f_unMessage, WPARAM f_wParam, LPARAM f_lParam )
{
  MKI_DUMMYSTRW szDummyStr;
  int           nDummyInt;

  switch( f_unMessage )
  {
    // Initialize dialog box
    case WM_INITDIALOG:
      MKI_DialogCenter( f_hDlg, nullptr );
      SetWindowText( f_hDlg, gDialogData.m_szCaption );

      // Convert the current indent number to a string that can be set to an edit field.
      MKI_PrintW( szDummyStr, L"%d", gDialogData.m_cCopyProperty.m_nIndentNum );
      SetDlgItemTextW( f_hDlg, IDD_EDIT0, szDummyStr );

      // Set the focus to the first edit field and highlight the complete content of that edit field.
      Edit_SetSel( GetDlgItem( f_hDlg, IDD_EDIT0 ), 0, -1 );

      gDialogData.m_cTooltip.SetTooltipList( f_hDlg, eTooltipDialogProperty );
      gDialogData.m_cSubclassList.AddDialogControls( f_hDlg );

      gDialogData.m_hOldFocus = SetFocus( GetDlgItem( f_hDlg, IDD_EDIT0 ) );
      return( false );

    case WM_COMMAND:
      switch( GET_WM_COMMAND_ID( f_wParam, f_lParam ) )
      {
        // OK
        case IDD_OK:
          // Read the indent number from the edit field and convert it to an integer value.
          MKI_DialogControlGetText( f_hDlg, IDD_EDIT0, szDummyStr );
          if( !MKI_ConvertStringInt32( nDummyInt, szDummyStr ) )
          {
            MKI_DialogMessage( eDialogText[3], eDialogText[0], MKI_MESSAGE_OK );

            Edit_SetSel( GetDlgItem( f_hDlg, IDD_EDIT0 ), 0, -1 );
            SetFocus( GetDlgItem( f_hDlg, IDD_EDIT0 ) );
            return( true );
          }

          // Check whether the stated indent number is within range. If not, display a message and
          // set the focus to the according edit field.
          if( nDummyInt < STAR_INDENT_MIN || nDummyInt > STAR_INDENT_MAX )
          {
            MKI_DialogMessage( eDialogText[3], eDialogText[0], MKI_MESSAGE_OK );

            Edit_SetSel( GetDlgItem( f_hDlg, IDD_EDIT0 ), 0, -1 );
            SetFocus( GetDlgItem( f_hDlg, IDD_EDIT0 ) );
            return( true );
          }
          gDialogData.m_cCopyProperty.m_nIndentNum = (__int32) nDummyInt;
          gDialogData.m_cCopyProperty.SetAll();

          gDialogData.m_nExitDialog = IDD_OK;
          SetFocus( gDialogData.m_hOldFocus );
          EndDialog( f_hDlg, true );
          return( true );

        // Cancel
        case IDD_CANCEL:
          gDialogData.m_nExitDialog = IDD_CANCEL;
          SetFocus( gDialogData.m_hOldFocus );
          EndDialog( f_hDlg, true );
          return( true );
      }
      break;

    case WM_HELP:
      MKI_DialogHelpTopic( f_hDlg, MK_PRODUCT_REGKEY, gGlobalData.m_nCommandID );
      return( true );

    default:
      return( MKI_DialogHandleDefaultMessages( f_hDlg, f_unMessage, f_wParam, f_lParam ) );
  }
  return( false );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure displays and manages the parameter dialog for the command "Star" that this plug-in offers to the user.
bool
DialogEditParameters( HWND f_hMainWnd, const LPCWSTR f_pszCaption, CPropertyData& f_rData )
{
  MKI_CopyW( gDialogData.m_szCaption, f_pszCaption );

  // Copy the current settings to the temporary copy in order to be able to ignore the changes if the user cancels the dialog.
  gDialogData.m_cCopyProperty = f_rData;

  // Call and process the dialog window.
  if( MKI_DialogBoxParam( gGlobalData.m_hLanguage, "EDITPARAM", f_hMainWnd, DialogEditParametersManage, (LPARAM) 0 ) == -1 )
  {
    MKI_DialogMessage( eDialogText[2], eDialogText[0], MKI_MESSAGE_OK );
    return( false );
  }

  // Check whether the user cancelled the dialog or not. If he pressed OK, copy the new values to
  // the current settings.
  if( gDialogData.m_nExitDialog == IDD_CANCEL )
    return( false );

  f_rData = gDialogData.m_cCopyProperty;
  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// Callback procedure for dialog box handling of DialogEditParameters__(). This procedure handles all messages sent to the dialog window.
INT_PTR CALLBACK
DialogEditParametersManage__( HWND f_hDlg, UINT f_unMessage, WPARAM f_wParam, LPARAM f_lParam )
{
  MKI_DUMMYSTRW szDummyStr;
  int           nDummyInt;

  switch( f_unMessage )
  {
    // Initialize dialog box
    case WM_INITDIALOG:
      MKI_DialogCenter( f_hDlg, nullptr );
      SetWindowText( f_hDlg, gDialogData.m_szCaption );

      // Convert the current indent number to a string that can be set to an edit field.
      gDialogData.m_fUserUpdate = false;
      if( gDialogData.m_cCopyProperty.IsIndeterminate( CPropertyData::STAR_USE_INDENTNUM ) )
      {
        SetDlgItemTextW( f_hDlg, IDD_EDIT0, L"" );
      }
      else
      {
        MKI_PrintW( szDummyStr, L"%d", gDialogData.m_cCopyProperty.m_nIndentNum );
        SetDlgItemTextW( f_hDlg, IDD_EDIT0, szDummyStr );
      }

      MKI_DialogInitCheck( f_hDlg, IDD_CHECK0, gDialogData.m_cCopyProperty.m_nUse & CPropertyData::STAR_USE_INDENTNUM );
      gDialogData.m_fUserUpdate = true;

      gDialogData.m_cTooltip.SetTooltipList( f_hDlg, eTooltipDialogProperty );
      gDialogData.m_cSubclassList.AddDialogControls( f_hDlg );

      // Set the focus to the first edit field and highlight the complete content of that edit field.
      Edit_SetSel( GetDlgItem( f_hDlg, IDD_EDIT0 ), 0, -1 );
      gDialogData.m_hOldFocus = SetFocus( GetDlgItem( f_hDlg, IDD_EDIT0 ) );
      return( false );

    case WM_COMMAND:
      switch( GET_WM_COMMAND_ID( f_wParam, f_lParam ) )
      {
        // USE_INDENTNUM
        case IDD_CHECK0:
          MKI_DialogHandleCheckInvert( f_hDlg, IDD_CHECK0, gDialogData.m_cCopyProperty.m_nUse, CPropertyData::STAR_USE_INDENTNUM );
          return( true );

        case IDD_EDIT0:
          if( GET_WM_COMMAND_CMD( f_wParam, f_lParam ) == EN_CHANGE && gDialogData.m_fUserUpdate )
            MKI_DialogHandleCheckSet( f_hDlg, IDD_CHECK0, gDialogData.m_cCopyProperty.m_nUse, CPropertyData::STAR_USE_INDENTNUM );
          return( true );

        // OK
        case IDD_OK:
          if( !gDialogData.m_cCopyProperty.IsIndeterminate( CPropertyData::STAR_USE_INDENTNUM ) )
          {
            // Read the indent number from the edit field and convert it to an integer value.
            MKI_DialogControlGetText( f_hDlg, IDD_EDIT0, szDummyStr );

            // Check whether the stated indent number is within range. If not, display a message and
            // set the focus to the according edit field.
            if( !MKI_ConvertStringInt32( nDummyInt, szDummyStr ) ||
                nDummyInt < STAR_INDENT_MIN || nDummyInt > STAR_INDENT_MAX )
            {
              MKI_DialogMessage( eDialogText[3], eDialogText[0], MKI_MESSAGE_OK );
              Edit_SetSel( GetDlgItem( f_hDlg, IDD_EDIT0 ), 0, -1 );
              SetFocus( GetDlgItem( f_hDlg, IDD_EDIT0 ) );
              return( true );
            }
            gDialogData.m_cCopyProperty.m_nIndentNum = (__int32) nDummyInt;
          }

          gDialogData.m_nExitDialog = IDD_OK;
          SetFocus( gDialogData.m_hOldFocus );
          EndDialog( f_hDlg, true );
          return( true );

        case IDD_CANCEL:
          gDialogData.m_nExitDialog = IDD_CANCEL;
          SetFocus( gDialogData.m_hOldFocus );
          EndDialog( f_hDlg, true );
          return( true );
      }
      break;

    case WM_HELP:
      MKI_DialogHelpTopic( f_hDlg, MK_PRODUCT_REGKEY, gGlobalData.m_nCommandID );
      return( true );

    default:
      return( MKI_DialogHandleDefaultMessages( f_hDlg, f_unMessage, f_wParam, f_lParam ) );
  }
  return( false );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure displays and manages the parameter dialog for the command "Star" that this plug-in offers to the user.
bool
DialogEditParameters__( HWND f_hMainWnd, const LPCWSTR f_pszCaption, CPropertyData& f_rData )
{
  MKI_CopyW( gDialogData.m_szCaption, f_pszCaption );

  // Copy the current settings to the temporary copy in order to be able to ignore the changes if the user cancels the dialog.
  gDialogData.m_cCopyProperty = f_rData;

  // Call and process the dialog window.
  if( MKI_DialogBoxParam( gGlobalData.m_hLanguage, "EDITPARAM__", f_hMainWnd, DialogEditParametersManage__, (LPARAM) 0 ) == -1 )
  {
    MKI_DialogMessage( eDialogText[2], eDialogText[0], MKI_MESSAGE_OK );
    return( false );
  }

  // Check whether the user cancelled the dialog or not. If he pressed OK, copy the new values to
  // the current settings.
  if( gDialogData.m_nExitDialog == IDD_CANCEL )
    return( false );

  f_rData = gDialogData.m_cCopyProperty;
  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is used to calculate the polygon corner points of a star.
__int32
CreateStarPolyline( MKI_CONST_POINT_REF f_rPoint1, MKI_CONST_POINT_REF f_rPoint2, 
                    MKI_CONST_POINT_REF f_rPoint3, MKI_CONST_POINT_REF f_rPoint4,
                    __int32 f_nIndentNum, MKI_POINT_PTR f_pResult )
{
  double  dRadius1, 
          dRadius2, 
          dRotation, 
          dAngle1, 
          dAngle2,
          dStep;
  __int32 nIndex;

  // Calculation of rotation angle (distance p1 - p4)
  dRotation = f_rPoint1.AngleTo( f_rPoint4 );

  // Calculation of the first radius (distance p1 - p2)
  dRadius1 = f_rPoint1.LengthTo( f_rPoint2 );

  // Calculation of the second radius (distance p1 - p3)
  dRadius2 = f_rPoint1.LengthTo( f_rPoint3 );

  // Step angle
  if( f_nIndentNum > 0 )
    dStep = MKI_MATH_2PI / f_nIndentNum;
  else
    dStep = MKI_MATH_PI;

  // Calculation of point pairs with an offset of 1/2 step
  dAngle1 = dRotation;
  dAngle2 = dRotation + 0.5 * dStep;

  nIndex = 0;
  for( __int32 nCount = 0; nCount < f_nIndentNum; nCount++ )
  {
    f_pResult[nIndex  ].x = f_rPoint1.x + cos( dAngle1 ) * dRadius2;
    f_pResult[nIndex++].y = f_rPoint1.y + sin( dAngle1 ) * dRadius2;
    f_pResult[nIndex  ].x = f_rPoint1.x + cos( dAngle2 ) * dRadius1;
    f_pResult[nIndex++].y = f_rPoint1.y + sin( dAngle2 ) * dRadius1;
    dAngle1 += dStep;
    dAngle2 += dStep;
  }

  // Number of calculated points
  return( nIndex );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure creates an object of types "Surface" and fills it with required definition points to build a star.
bool
FinishCommandStar( MKI_CONST_POINT_REF f_rPoint1, MKI_CONST_POINT_REF f_rPoint2, 
                   MKI_CONST_POINT_REF f_rPoint3, MKI_CONST_POINT_REF f_rPoint4,
                   __int32 f_nIndentNum, MKI_CONST_XPROPERTY_PTR f_pXProperty, MKI_XPROPERTY_PTR f_pNewXProperty )
{
  MKI_POINT aPoly[STAR_INDENT_MAX*2];
  bool      fResult = false;

  // Fill the array aPoly[] with a list of point coordinates
  int nTotal = CreateStarPolyline( f_rPoint1, f_rPoint2, f_rPoint3, f_rPoint4, f_nIndentNum, aPoly );

  // Open an object of "Surface"
  if( MKI_ENTITY_OBJECT_PTR pObj = MKI_ObjectOpen( MKI_OBJ_SURFACE ) )
  {
    if( f_pNewXProperty )
      *f_pNewXProperty = pObj->m_cXProperty;
    if( f_pXProperty )
      pObj->m_cXProperty = *f_pXProperty;

    // Add the start point of the polygon
    MKI_ObjectAddPoint( MKI_DB_POINT_START, aPoly[0] );

    // Add all further points as end point
    for( int nCount = 1; nCount < nTotal; nCount++ )
      MKI_ObjectAddPoint( MKI_DB_POINT_END, aPoly[nCount] );

    // Close the object, initialize its properties and insert it into the current drawing
    if( MKI_ObjectFastInsert( nullptr ) )
      fResult = true;
    else
      MKI_FeedbackNotification( MKI_NOTIFICATIONMODE_BLINKING, MKI_NOTIFICATIONTIME_LONG, eDialogText[2], nullptr );
  }
  else
    MKI_FeedbackNotification( MKI_NOTIFICATIONMODE_BLINKING, MKI_NOTIFICATIONTIME_LONG, eDialogText[2], nullptr );

  return( fResult );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure creates an object of types "Surface" and fills it with required definition points to build a star.
bool
FinishCommandExtStar( MKI_CONST_POINT_REF f_rPoint1, MKI_CONST_POINT_REF f_rPoint2, 
                      MKI_CONST_POINT_REF f_rPoint3, MKI_CONST_POINT_REF f_rPoint4,
                      __int32 f_nIndentNum )
{
  bool  fResult = false;

  // Create an extended object referencing the previously created group
  if( MKI_ENTITY_EXT_PTR pExtObj = MKI_ExtOpen( MKI_DB_OWNER_MK, MKI_OBJ_SAMPLE_STAR, MKI_EXTFLAG_CURVE | MKI_EXTFLAG_SURFACE, SAMPLE_PLUGIN_VERSION ) )
  {
    MKI_XPROPERTY cXProperty;

    if( !FinishCommandStar( f_rPoint1, f_rPoint2, f_rPoint3, f_rPoint4, f_nIndentNum, nullptr, &cXProperty ) )
      fResult = false;

    pExtObj->m_cXProperty = cXProperty;
    pExtObj->m_nDisplayMode = MKI_DISPLAYMODE_NOLISTPOINTS;

    // Add all data that is required to completely regenerate the star
    MKI_ExtAddPoint( MKI_DB_POINT_CENTER, f_rPoint1 );
    MKI_ExtAddPoint( MKI_DB_POINT_RADIUS, f_rPoint2 );
    MKI_ExtAddPoint( MKI_DB_POINT_RADIUS, f_rPoint3 );
    MKI_ExtAddPoint( MKI_DB_POINT_ANGLE,  f_rPoint4 );
    MKI_ExtAddInt32( 1000, f_nIndentNum );

    if( MKI_ExtFastInsert( nullptr ) )
      fResult = true;
    else
      MKI_FeedbackNotification( MKI_NOTIFICATIONMODE_BLINKING, MKI_NOTIFICATIONTIME_LONG, eDialogText[2], nullptr );
  }
  else
    MKI_FeedbackNotification( MKI_NOTIFICATIONMODE_BLINKING, MKI_NOTIFICATIONTIME_LONG, eDialogText[2], nullptr );

  return( fResult );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called by the serving application during the execution of the plug-in-supplied command "Star". It is called each time the entry
// of a new point is about to start.
__int32
MKI_InputPointInitProc( __int32 f_nCommandID, __int32 f_nPointIndex, MKI_CONST_POINT_REF f_rPoint )
{
  f_nCommandID;   // Touch formal parameter to avoid C4100
  f_nPointIndex;  // Touch formal parameter to avoid C4100
  f_rPoint;       // Touch formal parameter to avoid C4100

  // There is no need to initialize anything, so return immediately.

  return( MKI_INPUT_OK );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called by the serving application during the execution of the plug-in-supplied command "Star". It is called at least once each
// time the user moves the cursor.
bool
MKI_InputPointMoveProc( __int32 f_nCommandID, __int32 f_nPointIndex, MKI_CONST_POINT_REF f_rPoint )
{
  f_nCommandID; // Touch formal parameter to avoid C4100

  // The only action required is to store the current point's coordinates. These coordinates will be used
  // to draw the "rubber lines" within the procedure MKI_InputDisplayProc() and later on to create the
  // object within the procedure MKI_InputFinishProc().
  // It will suffice if the coordinates are only stored at this one location since MKI_InputPointMoveProc()
  // will at least be called once with the final coordinate of each point.

  gGlobalData.m_acPoint[f_nPointIndex] = f_rPoint;

  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called by the serving application during the execution of the plug-in-supplied command "Star". It is called each time the entry
// of a point is to be finished.
__int32
MKI_InputPointExitProc( __int32 f_nCommandID, __int32 f_nPointIndex, MKI_CONST_POINT_REF f_rPoint )
{
  f_nCommandID;   // Touch formal parameter to avoid C4100
  f_nPointIndex;  // Touch formal parameter to avoid C4100
  f_rPoint;       // Touch formal parameter to avoid C4100

  // There is no need to do anything, so return immediately. The final point coordinates that are passed
  // in f_dX and f_dY should have already been saved in the MKI_InputPoitMoveProc(), so there is no need
  // to do it here again.

  return( MKI_INPUT_OK );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called by the serving application during the execution of the plug-in-supplied command "Star". It is called each time the
// plug-in has to update the "rubber lines" displaying the current input status on the screen.
void
MKI_InputDisplayProc( __int32 f_nCommandID, __int32 f_nPointIndex, MKI_DISPLAYDATA_PTR f_pData )
{
  f_pData;  // Touch formal parameter to avoid C4100

  // Room to store the polygon points of the calculated star.
  MKI_POINT   aPoly[STAR_INDENT_MAX*2];
  __int32     nTotal, 
              nCount;

  switch( f_nCommandID )
  {
    // Star
    case SAMPLE_MENU_POLYGON:
    case SAMPLE_MENU_EXTENDED:
      // Point index 0: No rubber lines visible during entry of first point (center).
      if( f_nPointIndex == 0 )
      {
        ;
      }

      // Point index 1: Display a circle defined by point index 0 (center) and point index 1 (radius 1)
      else if( f_nPointIndex == 1 )
      {
        MKI_InputDrawCircle( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[1] );
      }

      // Point index 2: Calculate and display a complete star based on point index 0 (center), point
      //                index 1 (radius 1) and point index 2 (radius 2).
      else if( f_nPointIndex == 2 )
      {
        MKI_InputDrawCircle( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[1] );
        MKI_InputDrawCircle( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[2] );

        nTotal = CreateStarPolyline( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[1], 
                                     gGlobalData.m_acPoint[2], gGlobalData.m_acPoint[0], gUserData.m_nIndentNum, aPoly );

        for( nCount = 1; nCount < nTotal; nCount++ )
          MKI_InputDrawLine( aPoly[nCount-1], aPoly[nCount] );

        MKI_InputDrawLine( aPoly[nTotal-1], aPoly[0] );
      }

      // Point index 3: Calculate and display a complete star based on point index 0 (center), point
      //                index 1 (radius 1), point index 2 (radius 2) and point index 3 (angle)
      else if( f_nPointIndex == 3 )
      {
        MKI_InputDrawCircle( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[1] );
        MKI_InputDrawCircle( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[2] );
        MKI_InputDrawLine  ( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[3] );

        nTotal = CreateStarPolyline( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[1], 
                                     gGlobalData.m_acPoint[2], gGlobalData.m_acPoint[3], gUserData.m_nIndentNum, aPoly );

        for( nCount = 1; nCount < nTotal; nCount++ )
          MKI_InputDrawLine( aPoly[nCount-1], aPoly[nCount] );

        MKI_InputDrawLine( aPoly[nTotal-1], aPoly[0] );
      }
      break;
  }
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called by the serving application during the execution of the plug-in-supplied command "Star". It is called each time the user
// wants to edit the parameters of the currently active command.
bool
MKI_InputParameterProc( __int32 f_nCommandID, __int32 f_nMode )
{
  gGlobalData.m_nCommandID = f_nCommandID;

  switch( f_nCommandID )
  {
    // Star
    case SAMPLE_MENU_POLYGON:
    case SAMPLE_MENU_EXTENDED:
      if( f_nMode ) 
      {
        // Display the parameter dialog window. If the parameters are changed (i.e. if DialogEditParameters()
        // returns true), do also return true to indicate the serving application that parameters of the
        // currently active command have changed, i.e. this command should be re-initialised.
        return( DialogEditParameters( gGlobalData.m_hGlobalWnd, eCommandName[0], gUserData ) );
      }
      // Return true to indicate that a parameter dialog is available for this command.
      return( true );
  }
  return( false );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called by the serving application during the execution of the plug-in-supplied command "Star". It is called each time the user
// explicity cancels the currently active command by pressing the right mouse button. If the current command were a command that requires a non-fixed
// number of point entries, this is the place to terminate the command by returning MKI_INPUT_FINISH. If the current command allows parameters to be
// edited DURING its execution (like the application's DRAW > CURVE command), this is the right place to display a corresponding dialog window.
__int32
MKI_InputCancelProc( __int32 f_nCommandID, __int32 f_nPointCount )
{
  f_nCommandID;   // Touch formal parameter to avoid C4100
  f_nPointCount;  // Touch formal parameter to avoid C4100

  return( MKI_INPUT_CANCEL );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called by the serving application during the execution of the plug-in-supplied command "Star". It is called each time the user
// has successfully enteres all required points, i.e. the command is finished and the resulting actions are to be performed.
void
MKI_InputFinishProc( __int32 f_nCommandID, __int32 f_nPointCount )
{
  f_nPointCount;  // Touch formal parameter to avoid C4100

  bool  fResult = true;

  // Start a new undo level. All objects created since now can either now be inserted to the drawing by
  // calling MKI_UndoFinishProcess() or can be deleted by calling MKI_UndoCancelProcess().
  MKI_UndoInitProcess();

  // Execute the command, i.e. calculate the resulting polygon and create an object of type "Surface"
  // based on that calculation.
  switch( f_nCommandID )
  {
    // Star
    case SAMPLE_MENU_POLYGON:
      if( !FinishCommandStar( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[1], 
                              gGlobalData.m_acPoint[2], gGlobalData.m_acPoint[3], gUserData.m_nIndentNum, nullptr, nullptr ) )
        fResult = false;
      break;

    // Star
    case SAMPLE_MENU_EXTENDED:
      if( !FinishCommandExtStar( gGlobalData.m_acPoint[0], gGlobalData.m_acPoint[1], 
                                 gGlobalData.m_acPoint[2], gGlobalData.m_acPoint[3], gUserData.m_nIndentNum ) )
        fResult = false;
      break;
  }

  // If creation was successful, end the newly created undo level (which inserts all objects to the
  // current drawing) and draw all newly created objects. Else, delete all newly created objects by
  // canceling the undo level.
  if( fResult )
  {
    MKI_UndoFinishProcess( MKI_CHANGECODE_ADD );
    MKI_UndoUpdateLinks();
    MKI_DrawNewObjects();
  }
  else
    MKI_UndoCancelProcess();
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This DLL entry procedure must exist in any DLL to be used in Win32. Since our DLL does all necessary initialization in its MKI_PlugInInit()
// procedure, this procedure is quite empty.
BOOL WINAPI
DllMain( HINSTANCE f_hInstance, DWORD f_dwReason, LPVOID f_pDummy )
{
  f_pDummy; // Touch formal parameter to avoid C4100

  switch( f_dwReason )
  {
    case DLL_PROCESS_ATTACH:
      gGlobalData.m_hInstDLL = f_hInstance;
      break;

    case DLL_PROCESS_DETACH:
      gGlobalData.m_hInstDLL = nullptr;
      break;
  }
  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the plug-in is loaded by the serving application. Its main tasks are:
// - Checking whether it is compatible with the given f_nInterfaceVersion
// - Checking whether it is licensed to the given serial number (optional)
// - Storing of the serving application's instance and main windows handle for further use
// - Loading of the language-dependent library
// - Filling in the plug-in ID structure of which address is passed in f_pPlugInID
// - Loading of options from the registry database
// - Loading profiles
// - Allocating any static memory required
MKI_INTERFACE_EXPORT __int32
MKI_PlugInInit( const LPCSTR f_pszSerialNumber, HINSTANCE f_hMainInst, HWND f_hMainWnd,
                __int32 f_nInterfaceVersion, MKI_PLUGIN_ID_REF f_rPlugInID )
{
  f_pszSerialNumber;  // Touch formal parameter to avoid C4100

  MKI_FILENAMEW szFileName;

  if( f_nInterfaceVersion > MKI_INTERFACE_VERSION )
    return( MKI_INIT_OLD_PLUGIN );
  else if( f_nInterfaceVersion < MKI_INTERFACE_VERSION )
    return( MKI_INIT_OLD_APPLICATION );

  gGlobalData.m_hGlobalInst = f_hMainInst;
  gGlobalData.m_hGlobalWnd  = f_hMainWnd;

  // Load the command icon bitmap from the language DLL.
  gGlobalData.m_hLanguage = LoadLibraryW( MK_PRODUCT_REGKEY L".DLL" );
  MKI_FileApplicationPath( MK_PRODUCT_REGKEY L".PNG", szFileName );
  gGlobalData.m_pBitmap = MKI_BitmapLoadImage( szFileName );

  // Fill the plug-in information data block
  f_rPlugInID.m_nOwnerID    = MKI_DB_OWNER_MK;
  f_rPlugInID.m_nPlugInID   = SAMPLE_PLUGIN_ID;
  f_rPlugInID.m_nPlugInCTRL = MKI_PLUGINCTRL_STARTER | MKI_PLUGINCTRL_PRO | MKI_PLUGINCTRL_TEST | MKI_PLUGINCTRL_DEMO | MKI_PLUGINCTRL_READONLY;
  f_rPlugInID.m_pContainer  = nullptr;

  f_rPlugInID.m_cPlugInProc.m_pfnInputPointInitProc = (MKI_INPUTPOINTINIT_PROC) MKI_InputPointInitProc;
  f_rPlugInID.m_cPlugInProc.m_pfnInputPointMoveProc = (MKI_INPUTPOINTMOVE_PROC) MKI_InputPointMoveProc;
  f_rPlugInID.m_cPlugInProc.m_pfnInputPointExitProc = (MKI_INPUTPOINTEXIT_PROC) MKI_InputPointExitProc;
  f_rPlugInID.m_cPlugInProc.m_pfnInputDisplayProc   = (MKI_INPUTDISPLAY_PROC)   MKI_InputDisplayProc;
  f_rPlugInID.m_cPlugInProc.m_pfnInputParameterProc = (MKI_INPUTPARAMETER_PROC) MKI_InputParameterProc;
  f_rPlugInID.m_cPlugInProc.m_pfnInputCancelProc    = (MKI_INPUTCANCEL_PROC)    MKI_InputCancelProc;
  f_rPlugInID.m_cPlugInProc.m_pfnInputFinishProc    = (MKI_INPUTFINISH_PROC)    MKI_InputFinishProc;

  f_rPlugInID.m_cPlugInData.m_nType = MKI_PLUGINMENU_DRAW;

  f_rPlugInID.m_cPlugInData.m_cInputData.m_nCommandMode = MKI_COMMAND_DIRECT;
  f_rPlugInID.m_cPlugInData.m_cMenuData.m_pszMenuEntry   = eStartUpText[1];
  f_rPlugInID.m_cPlugInData.m_cMenuData.m_pszDescription = eStartUpText[2];

  f_rPlugInID.m_cPlugInData.m_cIconData.m_pIconBitmap  = gGlobalData.m_pBitmap;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconXOffset = MKI_ICON_COMMAND_FRAME;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconYOffset = MKI_ICON_COMMAND_FRAME;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconXSize   = MKI_ICON_COMMAND_SIZE;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconYSize   = MKI_ICON_COMMAND_SIZE;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconMode    = 0;

  f_rPlugInID.m_pCommandData  = gGlobalData.m_acCommandData;
  f_rPlugInID.m_pButtonData   = nullptr;
  f_rPlugInID.m_pExtendedData = gGlobalData.m_acExtendedData;

  // Fill in the commands' names and the accompanying information.
  int nCount = 0;
  while( eCommandEntry[nCount] == nullptr || eCommandEntry[nCount][0] != MKI_END_CHAR )
  {
    gGlobalData.m_acCommandData[nCount].m_cInputData.m_nCommandMode = MKI_COMMAND_DIRECT;
    gGlobalData.m_acCommandData[nCount].m_cMenuData.m_pszMenuEntry   = eCommandEntry[nCount];
    gGlobalData.m_acCommandData[nCount].m_cMenuData.m_pszDescription = eCommandName [nCount];

    gGlobalData.m_acCommandData[nCount].m_cIconData.m_pIconBitmap  = gGlobalData.m_pBitmap;
    gGlobalData.m_acCommandData[nCount].m_cIconData.m_nIconXOffset = MKI_ICON_COMMAND_FRAME;
    gGlobalData.m_acCommandData[nCount].m_cIconData.m_nIconYOffset = MKI_ICON_COMMAND_FRAME + ( nCount + 1 ) * MKI_ICON_COMMAND_OFFSET;
    gGlobalData.m_acCommandData[nCount].m_cIconData.m_nIconXSize   = MKI_ICON_COMMAND_SIZE;
    gGlobalData.m_acCommandData[nCount].m_cIconData.m_nIconYSize   = MKI_ICON_COMMAND_SIZE;
    gGlobalData.m_acCommandData[nCount].m_cIconData.m_nIconMode    = 0;

    if( eCommandEntry[nCount] )
      gGlobalData.m_acCommandData[nCount].m_nType = MKI_MENUMODE_ENTRY;
    else
      gGlobalData.m_acCommandData[nCount].m_nType = MKI_MENUMODE_ENTRY | MKI_MENUMODE_SEPARATOR;

    nCount++;
  }
  gGlobalData.m_acCommandData[nCount].m_nType = MKI_MENUMODE_END;

  // Fill the command description structure with the predefined global structure gPointInfo.
  gGlobalData.m_acCommandData[0].m_cInputData = gPointInfo[0];
  gGlobalData.m_acCommandData[0].m_cInputData.m_apszPointNames[1] = eNewPoint[0];
  gGlobalData.m_acCommandData[0].m_cInputData.m_apszPointNames[2] = eNewPoint[1];

  gGlobalData.m_acCommandData[1].m_cInputData = gGlobalData.m_acCommandData[0].m_cInputData;

  // Fill the extended object description structure
  gGlobalData.m_acExtendedData[0].m_nOwnerID = MKI_DB_OWNER_MK;
  gGlobalData.m_acExtendedData[0].m_nExtType = MKI_OBJ_SAMPLE_STAR;
  gGlobalData.m_acExtendedData[0].m_pszExtTitle = eDialogText[1];
  gGlobalData.m_acExtendedData[1].m_nOwnerID = -1;

  // Read the current settings (set to their default values if not found)
  PlugInLoadSettings();

  return( MKI_INIT_OK );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the plug-in is removed by the serving application. Its main tasks are:
// - Checking whether anything is to be saved. If so, it should display a message information the user
//   about it and allowing him to save those changes.
// - Freeing of all statically allocated memory.
// If this procedure returns false, the serving application will not be able to terminate. So please, do only return false if shutting down the
// plug-in now would severely damage or destroy user data.
MKI_INTERFACE_EXPORT bool
MKI_PlugInExit( bool f_fSaveSettings )
{
  if( f_fSaveSettings )
    PlugInSaveSettings();

  MKI_BitmapFreeImage( gGlobalData.m_pBitmap );

  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when a extended object is altered. Each command plug-in that handles extended objects must contain such a procedure.
// When called, the plug-in should test whether the extended object pointed to by ExtObj was created by itself. If so, it must react on the request
// coded in f_nNotifyMode.
MKI_INTERFACE_EXPORT __int32
MKI_PlugInNotify( MKI_ENTITY_EXT_PTR f_pExt, __int32 f_nNotifyMode, void* f_pData )
{
  bool  fResult = true;

  if( f_pExt )
  {
    // Is it an object generated by us?
    if( f_pExt->m_cHeader.m_nEntityOwner != MKI_DB_OWNER_MK )
    {
      return( MKI_NOTIFY_RESULT_UNKNOWN );
    }
    // Is it an object we can handle?
    if( f_pExt->m_nExtType != MKI_OBJ_SAMPLE_STAR )
    {
      return( MKI_NOTIFY_RESULT_UNKNOWN );
    }
    // Created by a newer version than we are?
    if( f_pExt->m_nVersion > SAMPLE_PLUGIN_VERSION )
    {
      return( MKI_NOTIFY_RESULT_UNKNOWN );
    }
  }

  switch( f_nNotifyMode )
  {
    // An undo-aware, multi-entity modification process (modify, copy, delete, etc.) was started
    case MKI_NOTIFY_MODIFY_INIT:
      // No linked resources or allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // The started undo-aware, multi-entity modification process was finished, react accordingly
    case MKI_NOTIFY_MODIFY_FINISH:
      // No linked resources or allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // The started undo-aware, multi-entity modification process was canceled, react accordingly
    case MKI_NOTIFY_MODIFY_CANCEL:
      // No linked resources or allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // Regenerate the display list based on the current data
    case MKI_NOTIFY_EXT_INIT:
      if( f_pExt )
      {
        if( MKI_ExtListOpen( f_pExt ) )
        {
          MKI_BLOCK_ENUM b1( f_pExt ),
                         b2( f_pExt ),
                         b3( f_pExt ),
                         b4( f_pExt ),
                         b5( f_pExt );
          b2.SetNext( b1 );
          b3.SetNext( b2 );
          b4.SetNext( b3 );
          b5.SetNext( b4 );

          if( !FinishCommandStar( b1.Point(), b2.Point(), b3.Point(), b4.Point(), b5.Int32()[0], &f_pExt->m_cXProperty, nullptr ) )
            fResult = false;

          MKI_ExtListClose( fResult );
        }
        else
          fResult = false;

        if( fResult )
          return( MKI_NOTIFY_RESULT_DONE );
      }
      break;

    // Set version information for future compatibility
    case MKI_NOTIFY_EXT_CHECK:
      if( f_pExt )
      {
        if( f_pExt->m_nVersion < SAMPLE_PLUGIN_VERSION )
        {
          f_pExt->m_nVersion = SAMPLE_PLUGIN_VERSION;
          return( MKI_NOTIFY_RESULT_DONE );
        }
        else if( f_pExt->m_nVersion > SAMPLE_PLUGIN_VERSION )
        {
          return( MKI_NOTIFY_RESULT_FAIL );
        }
      }
      return( MKI_NOTIFY_RESULT_IGNORE );

    // Extended object being created
    case MKI_NOTIFY_EXT_CREATE:
      // No linked resources or allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // Multiply with the matrix pointed to by Matrix
    case MKI_NOTIFY_EXT_MODIFY:
      if( f_pExt && f_pData )
        MKI_EditEntityPointsMatrix( (MKI_ENTITY_PTR) f_pExt, (MKI_CONST_MATRIX_PTR) f_pData );
      return( MKI_NOTIFY_RESULT_DONE );

    // Extended object being hidden
    case MKI_NOTIFY_EXT_HIDE:
      // No linked resources or allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // Extended object being displayed
    case MKI_NOTIFY_EXT_SHOW:
      // No linked resources or allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // Extended object being destroyed
    case MKI_NOTIFY_EXT_DESTROY:
      // No linked resources or allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // Return the title and content brief of the object
    case MKI_NOTIFY_EXT_GET_CONTENT:
      if( f_pExt && f_pData )
      {
        MKI_CopyW( (LPWSTR) f_pData, 256, eDialogText[1] );
        return( MKI_NOTIFY_RESULT_DONE );
      }
      break;

    // Return the static title of the object
    case MKI_NOTIFY_EXT_GET_TITLEPTR:
      if( f_pExt && f_pData )
      {
        *( (LPWSTR*) f_pData ) = eDialogText[1];
        return( MKI_NOTIFY_RESULT_DONE );
      }
      break;

    // Return additional information on the extended object (here: number of points)
    case MKI_NOTIFY_EXT_GET_INFORMATION:
      if( f_pExt && f_pData )
      {
        MKI_BLOCK_ENUM b1( f_pExt ),
                       b2( f_pExt ),
                       b3( f_pExt ),
                       b4( f_pExt ),
                       b5( f_pExt );
        b2.SetNext( b1 );
        b3.SetNext( b2 );
        b4.SetNext( b3 );
        b5.SetNext( b4 );

        MKI_DUMMYSTRW szDummyStr;
        MKI_PrintW( szDummyStr, eDialogText[4], b5.Int32()[0] );
        MKI_CopyW( (LPWSTR) f_pData, 256, szDummyStr );
        return( MKI_NOTIFY_RESULT_DONE );
      }
      break;

    // Return processed extended properties of the extended object
    case MKI_NOTIFY_EXT_GET_PROPERTY:
      // Use default behaviour of the application, nothing to do here
      break;

    // Determine available context commands
    case MKI_NOTIFY_EXT_CONTEXT_FILL:
      if( f_pExt && f_pData )
      {

/*
        MKI_CONTEXT_MENU_PTR pData = (MKI_CONTEXT_MENU_PTR) f_pData;
        pData->m_anCommandID[0] = SAMPLE_MENU_CONTEXT_1;
        pData->m_aszMenuText[0] = L"&1  Command Name 1";
        pData->m_anCommandID[1] = SAMPLE_MENU_CONTEXT_2;
        pData->m_aszMenuText[1] = L"&2  Command Name 2";
        pData->m_nCount = 2;
*/

        return( MKI_NOTIFY_RESULT_DONE );
      }
      break;

    // Apply chosen context command
    case MKI_NOTIFY_EXT_CONTEXT_APPLY:
      if( f_pExt && f_pData )
      {

/*
        MKI_CONTEXT_COMMAND_PTR pData = (MKI_CONTEXT_COMMAND_PTR) f_pData;
        switch( pData->m_nCommandID )
        {
          case SAMPLE_MENU_CONTEXT_1:
            // Apply command 1 to the object in f_pExt
            return( MKI_NOTIFY_RESULT_DONE );

          case SAMPLE_MENU_CONTEXT_2:
            // Apply command 2 to the object in f_pExt
            return( MKI_NOTIFY_RESULT_DONE );
        }
*/

        return( MKI_NOTIFY_RESULT_IGNORE );
      }
      break;

    // Show help on context command
    case MKI_NOTIFY_EXT_CONTEXT_HELP:
      if( f_pData )
      {

/*
        MKI_CONTEXT_COMMAND_PTR pData = (MKI_CONTEXT_COMMAND_PTR) f_pData;
        switch( pData->m_nCommandID )
        {
          case SAMPLE_MENU_CONTEXT_1:
          case SAMPLE_MENU_CONTEXT_2:
            MKI_DialogHelpTopic( g_hGlobalWnd, MK_PRODUCT_REGKEY, pData->m_nCommandID );
            return( MKI_NOTIFY_RESULT_DONE );
        }
*/

        return( MKI_NOTIFY_RESULT_IGNORE );
      }
      break;

    // Short text
    case MKI_NOTIFY_EXT_GET_SHORTTEXT:
      // No text contained, so nothing to do here
      return( MKI_NOTIFY_RESULT_IGNORE );

    // Find text
    case MKI_NOTIFY_FIND_TEXT:
      // No text contained, so nothing to do here
      return( MKI_NOTIFY_RESULT_IGNORE );

    // Edit text
    case MKI_NOTIFY_EDIT_TEXT_SINGLE:
    case MKI_NOTIFY_EDIT_TEXT_MULTI:
    case MKI_NOTIFY_VIEW_TEXT_MULTI:
      // No text contained, so nothing to do here
      return( MKI_NOTIFY_RESULT_IGNORE );

    // Create temporary display list
    case MKI_NOTIFY_LIST_START:
      // No temporary display list required, so nothing to do here
      return( MKI_NOTIFY_RESULT_IGNORE );

    // Discard temporary display list
    case MKI_NOTIFY_LIST_END:
      // No temporary display list required, so nothing to do here
      return( MKI_NOTIFY_RESULT_IGNORE );

    // Initalize property editing
    case MKI_NOTIFY_PROPERTY_INIT:
      gGlobalData.m_cProperty.SetNone();
      return( MKI_NOTIFY_RESULT_DONE );

    // Set all use flags
    case MKI_NOTIFY_PROPERTY_SETALL:
      if( (__int32) f_pData == MKI_OBJ_SAMPLE_STAR )
      {
        gGlobalData.m_cProperty.SetAll();
        return( MKI_NOTIFY_RESULT_DONE );
      }
      break;

    // Fill in property structure
    case MKI_NOTIFY_PROPERTY_FILL:
      if( f_pExt && f_pData )
      {
        MKI_MULTI_INFO_PTR pMultiInfo = (MKI_MULTI_INFO_PTR) f_pData;
        if( !pMultiInfo->m_acXProperty[0].m_fValid )
        {
          pMultiInfo->m_acXProperty[0].m_fValid = true;
          pMultiInfo->m_acXProperty[0].m_nUse = MKI_USE_NULL;
          pMultiInfo->m_nUseType = MKI_USETYPE_FIX;
        }
        pMultiInfo->m_acXProperty[0].Fill( f_pExt->m_cXProperty );

        MKI_BLOCK_ENUM b1( f_pExt ),
                       b2( f_pExt ),
                       b3( f_pExt ),
                       b4( f_pExt ),
                       b5( f_pExt );
        b2.SetNext( b1 );
        b3.SetNext( b2 );
        b4.SetNext( b3 );
        b5.SetNext( b4 );

        gGlobalData.m_cProperty.Fill( b5.Int32()[0] );
      }
      return( MKI_NOTIFY_RESULT_DONE );

    // Edit a single object's properties
    case MKI_NOTIFY_PROPERTY_SINGLE:
      if( DialogEditParameters( *( (HWND*) f_pData ), eDialogText[1], gGlobalData.m_cProperty ) )
        return( MKI_NOTIFY_RESULT_DONE );
      else
        return( MKI_NOTIFY_RESULT_IGNORE );

    // Edit multiple objects' properties
    case MKI_NOTIFY_PROPERTY_MULTIPLE:
      if( DialogEditParameters__( *( (HWND*) f_pData ), eDialogText[1], gGlobalData.m_cProperty ) )
        return( MKI_NOTIFY_RESULT_DONE );
      else
        return( MKI_NOTIFY_RESULT_IGNORE );

    // Apply property structure
    case MKI_NOTIFY_PROPERTY_APPLY:
      if( f_pExt && f_pData )
      {
        MKI_MULTI_INFO_PTR pMultiInfo = (MKI_MULTI_INFO_PTR) f_pData;
        pMultiInfo->m_acXProperty[0].Apply( f_pExt->m_cXProperty, false );

        MKI_BLOCK_ENUM b1( f_pExt ),
                       b2( f_pExt ),
                       b3( f_pExt ),
                       b4( f_pExt ),
                       b5( f_pExt );
        b2.SetNext( b1 );
        b3.SetNext( b2 );
        b4.SetNext( b3 );
        b5.SetNext( b4 );

        gGlobalData.m_cProperty.Apply( b5.Int32()[0] );
      }
      return( MKI_NOTIFY_RESULT_DONE );

    // Exit property editing
    case MKI_NOTIFY_PROPERTY_EXIT:
      // No allocated data, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );

    // Apply base properties structure
    case MKI_NOTIFY_PROPERTY_BASEAPPLY:
      // Let the application apply its default behaviour
      return( MKI_NOTIFY_RESULT_UNKNOWN );

    // Display feedback data
    case MKI_NOTIFY_DISPLAY_INIT:
    case MKI_NOTIFY_DISPLAY_EXT_START:
    case MKI_NOTIFY_DISPLAY_CALLBACK:
    case MKI_NOTIFY_DISPLAY_EXT_END:
    case MKI_NOTIFY_DISPLAY_EXIT:
      // No feedback data to be displayed, so nothing to do here
      return( MKI_NOTIFY_RESULT_DONE );
  }
  return( MKI_NOTIFY_RESULT_UNKNOWN );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when a plug-in's command is chosen by the user. For a command plug-in, its main tasks are:
// - If requested, display the help topic for the given command
// - Execute the chosen command (if it is a "direct" command that does not require point entry)
// - Register an external command (if it is a "indirect" command that does require point entry)
// - Error handling and display
MKI_INTERFACE_EXPORT __int32
MKI_PlugInCommand( __int32 f_nCommandID, __int32 f_nExecMode, void* f_pData )
{
  f_pData;  // Touch formal parameter to avoid C4100

  __int32 nResult = 0;

  // Check which type of action is requested
  switch( f_nExecMode )
  {
    // Display the corresponding help topic
    case MKI_EXECMODE_HELP:
      switch( f_nCommandID )
      {
        case SAMPLE_MENU_POLYGON:
        case SAMPLE_MENU_EXTENDED:
        case SAMPLE_MENU_ABOUT:
          MKI_DialogHelpTopic( gGlobalData.m_hGlobalWnd, MK_PRODUCT_REGKEY, gGlobalData.m_nCommandID );
          break;
      }
      break;

    // User selected a command
    case MKI_EXECMODE_USER:
    // System restarted a command
    case MKI_EXECMODE_SYSTEM:
      gGlobalData.m_nCommandID = f_nCommandID;
      switch( f_nCommandID )
      {
        case SAMPLE_MENU_POLYGON:
          // If the user has first selected the command from the menu, allow him to edit the parameters first.
          // If the command does only use seldomly-modified parameters, this is NOT necessary.
          if( f_nExecMode == MKI_EXECMODE_USER )
          {
            nResult = DialogEditParameters( gGlobalData.m_hGlobalWnd, eCommandName[0], gUserData );
          }
          break;

        case SAMPLE_MENU_EXTENDED:
          // If the user has first selected the command from the menu, allow him to edit the parameters first.
          // If the command does only use seldomly-modified parameters, this is NOT necessary.
          if( f_nExecMode == MKI_EXECMODE_USER )
          {
            nResult = DialogEditParameters( gGlobalData.m_hGlobalWnd, eCommandName[1], gUserData );
          }
          break;

        case SAMPLE_MENU_ABOUT:
          {
            MKI_PLUGIN_CONTAINER  cData;

            MKI_CopyW( cData.m_szRegistry, MK_PRODUCT_REGKEY );
            MKI_GetBuildDate( __DATE__, cData.m_szBuildDate );
            cData.m_pszCaption   = eDialogText[0];
            cData.m_pszVendor    = nullptr;
            cData.m_pszCopyright = eStartUpText[0];
            cData.m_nUserMode    = MKI_PLUGINMODE_FREE;

            MKI_DialogPlugInAbout( gGlobalData.m_hGlobalWnd, cData, nullptr );
          }
          nResult = 1;
          break;
      }
      break;

    // Return menu item state
    case MKI_EXECMODE_MENU_INIT:
      // Nothing to do here...
      break;

    // Read drawing-dependent data
    case MKI_EXECMODE_GET_PROFILE:
      // Nothing to do here...
      break;

    // Write drawing-dependent data
    case MKI_EXECMODE_SET_PROFILE:
      // Nothing to do here...
      break;

    // Undo selected by user
    case MKI_EXECMODE_UNDO:
      // Nothing to do here...
      break;

    // Redo selected by user
    case MKI_EXECMODE_REDO:
      // Nothing to do here...
      break;

    // Clear undo data
    case MKI_EXECMODE_CLEAR_UNDO:
      // Nothing to do here...
      break;
  }
  return( nResult );
}
